import typing
from enum import Enum

import wx


class ViewType(Enum):
    AUTO = 1
    GROUPS = 2
    TREE = 3


class ViewSelect:

    def __init__(self, frame: wx.Frame, auto_item: wx.MenuItem, groups_item: wx.MenuItem, tree_item: wx.MenuItem, change_callback: typing.Callable):
        self.__auto_item = auto_item
        self.__auto_item.Enable(False)
        self.__groups_item = groups_item
        self.__tree_item = tree_item
        self.__change_callback = change_callback
        frame.Bind(wx.EVT_MENU, self.event, auto_item)
        frame.Bind(wx.EVT_MENU, self.event, groups_item)
        frame.Bind(wx.EVT_MENU, self.event, tree_item)

        self.__auto_item.Check(False)
        self.__groups_item.Check(False)
        self.__tree_item.Check(True)

    def event(self, e: wx.Event):
        self.selected = e.GetId()
        self.__change_callback()

    @property
    def selected(self):
        return ViewType.AUTO if self.__auto_item.IsChecked() else ViewType.GROUPS if self.__groups_item.IsChecked() else ViewType.TREE

    @selected.setter
    def selected(self, selected):
        self.__auto_item.Check(selected == ViewType.AUTO)
        self.__groups_item.Check(selected == ViewType.GROUPS)
        self.__tree_item.Check(selected == ViewType.TREE)



import os
from typing import Callable, Optional, List, Union

import wx

from . import Environment


class FilePickerHandler:
    def __init__(self, file_dialog_factory: Callable, propagator: Optional[Callable], overwrite_question: bool = False):
        self.__file_dialog_factory = file_dialog_factory
        self.__propagator = propagator
        self.__overwrite_question = overwrite_question
        self.filename = None

    # noinspection PyUnusedLocal
    def on_menu_click(self, e: wx.Event):
        del e
        with self.__file_dialog_factory() as file_dialog:
            initial_dir = Environment.settings.get(file_dialog.GetMessage())
            if initial_dir is not None:
                file_dialog.SetPath(initial_dir + '/')
            if file_dialog.ShowModal() == wx.ID_CANCEL:
                return
            if hasattr(file_dialog, 'GetPaths'):
                filenames = file_dialog.GetPaths()
            else:
                filenames = [file_dialog.GetPath()]
            Environment.settings[file_dialog.GetMessage()] = os.path.dirname(filenames[0])
            filenames = filenames if len(filenames) > 1 else filenames[0]
            self.file_selected(filenames, file_dialog.GetParent())

    def file_selected(self, filename: Union[str, List[str]], window: wx.Window):
        use_filename = True
        if self.__overwrite_question and os.path.exists(filename):
            use_filename = False
            with wx.MessageDialog(window, f'File {os.path.basename(filename)} already exists.\n\nDo you want to replace it?',
                                  style=wx.YES | wx.NO | wx.NO_DEFAULT | wx.ICON_WARNING) as question_dialog:
                if question_dialog.ShowModal() == wx.ID_YES:
                    use_filename = True

        if use_filename:
            self.filename = filename
            if self.__propagator is not None:
                self.__propagator(filename)



import typing

import wx


class SingleFileDropTarget(wx.FileDropTarget):

    def __init__(self, callback: typing.Callable):
        wx.FileDropTarget.__init__(self)
        self.__callback = callback

    def OnDropFiles(self, _, __, filenames):
        if len(filenames) == 1:
            self.__callback(filenames[0])
            return True
        return False



import typing
from typing import Callable
from unittest import TestCase

from asn1editor.controller import Controller, Converter
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface


class TestValueInterface(ValueInterface):
    def register_change_event(self, callback: typing.Callable):
        pass

    def __init__(self):
        self.val = ''

    def get_value(self) -> str:
        return self.val

    def set_value(self, val: str):
        self.val = val


class TestOptionalInterface(OptionalInterface):
    def register_optional_event(self, callback: Callable):
        pass

    def __init__(self):
        self.val = False

    def get_has_value(self) -> bool:
        return self.val

    def set_has_value(self, val: bool):
        self.val = val


class TestValueController(TestCase):
    def test_add_controller(self):
        with self.assertRaises(Exception):
            root = Controller.RootController('root')
            controller = Controller.ValueController('test', root, TestValueInterface(), None, Converter.Str(0, ''))
            controller.add_controller('test', controller)

    def test_init(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()

        Controller.ValueController('test', root, value_interface, None, Converter.Str(0, 'default'))
        self.assertEqual(value_interface.val, 'default')

        Controller.ValueController('test', root, value_interface, None, Converter.Int(0, 12))
        self.assertEqual(value_interface.val, '12')

        Controller.ValueController('test', root, value_interface, None, Converter.Int(0, None))
        self.assertEqual(value_interface.val, '0')

        Controller.ValueController('test', root, value_interface, None, Converter.Int(12, None))
        self.assertEqual(value_interface.val, '12')

    def test_model_to_view(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()

        controller = Controller.ValueController('test', root, value_interface, None, Converter.Str(0, 'default'))
        controller.model_to_view({'test': 'new'})
        self.assertEqual(value_interface.val, 'new')

        controller.model_to_view('new2')
        self.assertEqual(value_interface.val, 'new2')

        with self.assertRaises(AssertionError):
            controller.model_to_view({})

        controller = Controller.ValueController('test', root, value_interface, optional_interface, Converter.Str(0, 'default'))
        controller.model_to_view({'test': 'new'})
        self.assertEqual(value_interface.val, 'new')
        self.assertTrue(optional_interface.val)

        controller.model_to_view({})
        self.assertFalse(optional_interface.val)

    def test_view_to_model(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()

        controller = Controller.ValueController('test', root, value_interface, optional_interface, Converter.Str(0, 'default'))
        value_interface.val = 'new'
        optional_interface.val = True
        self.assertEqual('new', controller.view_to_model())

        optional_interface.val = False
        self.assertIsNone(controller.view_to_model())


class TestBoolController(TestCase):
    def test_add_controller(self):
        with self.assertRaises(Exception):
            root = Controller.RootController('root')
            controller = Controller.BoolController('test', root, TestValueInterface(), None, True)
            controller.add_controller('test', controller)

    def test_init(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()

        Controller.BoolController('test', root, value_interface, None, True)
        self.assertTrue(value_interface.val)

        Controller.BoolController('test', root, value_interface, None, False)
        self.assertFalse(value_interface.val)

    def test_model_to_view(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()

        controller = Controller.BoolController('test', root, value_interface, None, True)
        controller.model_to_view({'test': 'false'})
        self.assertEqual(value_interface.val, 'false')

        controller.model_to_view('True')
        self.assertEqual(value_interface.val, 'True')

        with self.assertRaises(AssertionError):
            controller.model_to_view({})

        controller = Controller.BoolController('test', root, value_interface, optional_interface, True)
        controller.model_to_view({'test': 'false'})
        self.assertEqual(value_interface.val, 'false')
        self.assertTrue(optional_interface.val)

        controller.model_to_view({})
        self.assertFalse(optional_interface.val)

    def test_view_to_model(self):
        root = Controller.RootController('root')
        value_interface = TestValueInterface()
        optional_interface = TestOptionalInterface()

        controller = Controller.BoolController('test', root, value_interface, optional_interface, False)
        value_interface.val = 'True'
        optional_interface.val = True
        self.assertTrue(controller.view_to_model())

        optional_interface.val = False
        self.assertIsNone(controller.view_to_model())



import os
import typing

import wx

from asn1editor.ASN1SpecHandler import ASN1SpecHandler
from asn1editor.Plugin import Plugin
from asn1editor.wxPython import Resources
from asn1editor.wxPython.FilePickerHandler import FilePickerHandler
from asn1editor.wxPython.Resources import plugin_resource_path
from asn1editor.wxPython.ViewSelect import ViewSelect, ViewType


class MenuHandler:
    def __init__(self, frame: wx.Frame, plugins: typing.Optional[typing.List[Plugin]], about_box_content: str):
        self.__frame = frame
        self.__plugins = plugins
        self.__load_data_item = None
        self.__save_data_item = None
        self.__load_spec = None
        self.__load_last_spec = None
        self.__close_spec_item = None
        self.__recent: typing.Optional[typing.List[typing.List[str]]] = None
        self.__recent_menu: typing.Optional[wx.Menu] = None
        self.__about_box_content = about_box_content
        self.view_select: typing.Optional[ViewSelect] = None

    def build(self, load_spec: typing.Callable, load_data_from_file: typing.Callable, save_data_to_file: typing.Callable, view_changed: typing.Callable):
        self.__load_spec = load_spec

        menu_bar = wx.MenuBar()
        file_menu = wx.Menu()
        load_spec_item: wx.MenuItem = file_menu.Append(wx.ID_ANY, 'Open ASN.1 specification')
        load_spec_item.SetBitmap(Resources.get_bitmap_from_svg('open'))
        self.__close_spec_item = file_menu.Append(wx.ID_ANY, 'Close ASN.1 specification')
        self.__close_spec_item.Enable(False)
        self.__recent_menu = wx.Menu()
        self.__recent_menu.AppendSeparator()
        self.__frame.Bind(wx.EVT_MENU, self.__clear_recent, self.__recent_menu.Append(wx.ID_ANY, 'Clear recent list'))
        recent_submenu: wx.MenuItem = file_menu.AppendSubMenu(self.__recent_menu, 'Open recent')
        recent_submenu.SetBitmap(Resources.get_bitmap_from_svg('recent'))
        self.__load_last_spec: wx.MenuItem = file_menu.Append(wx.ID_ANY, 'Open last specification on startup', kind=wx.ITEM_CHECK)
        file_menu.AppendSeparator()
        self.__load_data_item: wx.MenuItem = file_menu.Append(wx.ID_OPEN, 'Load encoded data')
        self.__load_data_item.SetBitmap(Resources.get_bitmap_from_svg('load_encoded'))
        self.__load_data_item.Enable(False)
        self.__save_data_item: wx.MenuItem = file_menu.Append(wx.ID_SAVE, 'Save encoded data')
        self.__save_data_item.SetBitmap(Resources.get_bitmap_from_svg('save_encoded'))
        self.__save_data_item.Enable(False)
        file_menu.AppendSeparator()
        exit_item = file_menu.Append(wx.ID_EXIT, 'Exit', 'Exit application')
        exit_item.SetBitmap(Resources.get_bitmap_from_svg('exit'))
        menu_bar.Append(file_menu, '&File')

        view_menu = wx.Menu()
        auto_view: wx.MenuItem = view_menu.Append(ViewType.AUTO.value, 'Automatic', kind=wx.ITEM_CHECK)
        groups_view: wx.MenuItem = view_menu.Append(ViewType.GROUPS.value, 'Groups', kind=wx.ITEM_CHECK)
        tree_view: wx.MenuItem = view_menu.Append(ViewType.TREE.value, 'Tree', kind=wx.ITEM_CHECK)
        tree_view.Check(True)
        menu_bar.Append(view_menu, '&View')

        self.__frame.SetMenuBar(menu_bar)

        toolbar: typing.Optional[wx.ToolBar] = None

        if self.__plugins is not None:
            for plugin_index, plugin in enumerate(self.__plugins):
                plugin_menu = wx.Menu()
                menus = plugin.get_menus()
                for i, menu in enumerate(menus):
                    if not len(menu[0]):
                        plugin_menu.AppendSeparator()
                    else:
                        menu_item: wx.MenuItem = plugin_menu.Append(plugin_index * 1000 + i, menu[0])
                        if menu[1] is not None:
                            self.__frame.Bind(wx.EVT_MENU, self.__plugin_menu_event, menu_item)
                        else:
                            menu_item.Enable(False)

                menu_bar.Append(plugin_menu, plugin.get_name())

                tools = plugin.get_tools()
                if len(tools):
                    if toolbar is None:
                        toolbar = self.__frame.CreateToolBar()
                        toolbar.SetToolSeparation(8)
                        toolbar.Bind(wx.EVT_TOOL, self.__tb_menu_event)
                    else:
                        toolbar.AddSeparator()

                    labels = False

                    for i, tool in enumerate(tools):
                        if not len(tool):
                            toolbar.AddSeparator()
                        else:
                            bitmap = wx.Bitmap(plugin_resource_path(tool[2]))
                            labels |= len(tool[0])
                            toolbar.AddTool(toolId=plugin_index * 1000 + i, label=tool[0], bitmap=bitmap, shortHelp=tool[1])
                    if labels:
                        toolbar.SetWindowStyle(wx.TB_TEXT | wx.TB_HORIZONTAL)
                        pass

        if toolbar is not None:
            toolbar.Realize()

        help_menu = wx.Menu()
        about_item = help_menu.Append(wx.ID_ABOUT, 'About')
        self.__frame.Bind(wx.EVT_MENU, self.__about_item_event, about_item)
        menu_bar.Append(help_menu, '&Help')

        # Event binding
        def schema_dialog_constructor() -> wx.FileDialog:
            return wx.FileDialog(self.__frame, "ASN.1 schema", wildcard="ASN.1 files (*.asn)|*.asn",
                                 style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST | wx.FD_MULTIPLE)

        def data_load_dialog_constructor() -> wx.FileDialog:
            extensions = ';'.join(ASN1SpecHandler.get_extensions())
            return wx.FileDialog(self.__frame, "ASN.1 encoded file", wildcard=f"ASN.1 encoded ({extensions})|{extensions}",
                                 style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)

        def data_save_dialog_constructor() -> wx.FileDialog:
            extensions = ';'.join(ASN1SpecHandler.get_extensions())
            return wx.FileDialog(self.__frame, "ASN.1 encoded file", wildcard=f"ASN.1 encoded ({extensions})|{extensions}", style=wx.FD_SAVE)

        self.__frame.Bind(wx.EVT_MENU, self.__exit, exit_item)

        self.__frame.Bind(wx.EVT_MENU, self.__close_spec, self.__close_spec_item)

        picker = FilePickerHandler(schema_dialog_constructor, load_spec)
        self.__frame.Bind(wx.EVT_MENU, picker.on_menu_click, load_spec_item)

        picker = FilePickerHandler(data_load_dialog_constructor, load_data_from_file)
        self.__frame.Bind(wx.EVT_MENU, picker.on_menu_click, self.__load_data_item)

        picker = FilePickerHandler(data_save_dialog_constructor, save_data_to_file, True)
        self.__frame.Bind(wx.EVT_MENU, picker.on_menu_click, self.__save_data_item)

        self.view_select = ViewSelect(self.__frame, auto_view, groups_view, tree_view, view_changed)

    def enable(self, enable: bool = True):
        self.__load_data_item.Enable(enable)
        self.__save_data_item.Enable(enable)
        self.__close_spec_item.Enable(enable)

    @property
    def recent(self) -> typing.List[typing.List[str]]:
        return self.__recent

    @recent.setter
    def recent(self, recent_list: typing.List[typing.List[str]]):
        self.__recent = recent_list

        for recent_list in reversed(self.__recent):
            self.__prepend_recent_to_menu(recent_list)

    def add_recent(self, filename: str, typename: str):
        recent = [filename, typename]
        if recent not in self.__recent:
            self.__recent.insert(0, recent)
            self.__prepend_recent_to_menu(recent)

    def __prepend_recent_to_menu(self, recent: typing.List[str]):
        recent_item = self.__recent_menu.Prepend(wx.ID_ANY, f'{os.path.basename(recent[0])} {recent[1]} ({recent[0]})')
        self.__frame.Bind(wx.EVT_MENU, lambda _: self.__load_spec(recent[0], recent[1]), recent_item)

    def __clear_recent(self, _: wx.Event):
        menu_items = self.__recent_menu.GetMenuItems()
        for menu_item in menu_items:
            if menu_item.GetKind() == wx.ITEM_SEPARATOR:
                break
            self.__recent_menu.RemoveItem(menu_item)

        self.__recent = []

    def load_most_recent(self):
        if len(self.__recent):
            most_recent = self.__recent[0]
            self.__load_spec(most_recent[0], most_recent[1])

    @property
    def load_last(self) -> bool:
        return self.__load_last_spec.IsChecked()

    @load_last.setter
    def load_last(self, load_last: bool):
        self.__load_last_spec.Check(load_last)

    # noinspection PyUnusedLocal
    def __about_item_event(self, e: wx.Event):
        del e
        dialog = wx.MessageDialog(self.__frame, self.__about_box_content, style=wx.ICON_INFORMATION | wx.OK, caption='About')
        dialog.ShowModal()

    def __tb_menu_event(self, e):
        menu_id = e.GetId()
        plugin_index = menu_id // 1000
        self.__plugins[plugin_index].get_tools()[menu_id % 1000][3]()

    def __plugin_menu_event(self, e):
        menu_id = e.GetId()
        plugin_index = menu_id // 1000
        self.__plugins[plugin_index].get_menus()[menu_id % 1000][1]()

    # noinspection PyUnusedLocal
    def __exit(self, e: wx.Event):
        del e
        self.__frame.Close()

    def __close_spec(self, _):
        self.enable(False)
        self.__load_spec(None)



import json
import os
import typing


class Styler:
    def __init__(self, style_file_name: str):
        self.__styles = {}
        if os.path.exists(style_file_name):
            with open(style_file_name) as f:
                self.__styles = json.load(f)

    def get_style(self, name: str) -> typing.Optional[str]:
        return self.__styles.get(name)



from typing import Tuple, List, Optional

from asn1editor.interfaces.BitstringInterface import BitstringInterface
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface
from asn1editor.view.AbstractView import AbstractView, ContainerView, ListView, ChoiceView


class AbstractViewFactory:  # pragma: no cover
    def get_number_view(self, name: str, optional: bool, minimum: int, maximum: int, float_: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        raise NotImplementedError

    def get_text_view(self, name: str, text: str) -> AbstractView:
        raise NotImplementedError

    def get_container_view(self, name: str, optional: bool) -> Tuple[ContainerView, OptionalInterface]:
        raise NotImplementedError

    def get_list_view(self, name: str, minimum: int, maximum: int, optional: bool) -> Tuple[ListView, ValueInterface, OptionalInterface]:
        raise NotImplementedError

    def get_enumerated_view(self, name: str, choices: List[str], optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        raise NotImplementedError

    def get_boolean_view(self, name: str, optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        raise NotImplementedError

    def get_hex_string_view(self, name: str, optional: bool, minimum: Optional[int], maximum: Optional[int]) -> \
            Tuple[ChoiceView, ValueInterface, OptionalInterface]:
        raise NotImplementedError

    def get_string_view(self, name: str, string_type: str, optional: bool, minimum: int, maximum: int) -> Tuple[ChoiceView, ValueInterface, OptionalInterface]:
        raise NotImplementedError

    def get_choice_view(self, name: str, choices: List[str], optional: bool) -> Tuple[ChoiceView, ValueInterface, OptionalInterface]:
        raise NotImplementedError

    def get_bitstring_view(self, name: str, number_of_bits: int, named_bits: List[Tuple[int, str]], optional: bool) -> Tuple[ChoiceView, BitstringInterface,
                                                                                                                             OptionalInterface]:
        raise NotImplementedError



import typing


class BitstringInterface:  # pragma: no cover
    def get_values(self) -> typing.List[int]:
        raise NotImplementedError

    def set_values(self, values: typing.List[int]):
        raise NotImplementedError



import os
import sys
import typing

import wx
import wx.svg

import asn1editor
from asn1editor.ASN1SpecHandler import ASN1SpecHandler
from asn1editor.Plugin import Plugin
from asn1editor.PluginInterface import PluginInterface
from asn1editor.wxPython import Environment, Resources
from asn1editor.wxPython import WxPythonViewFactory
from asn1editor.wxPython.FilePickerHandler import FilePickerHandler
from asn1editor.wxPython.ImageList import ImageList
from asn1editor.wxPython.MenuHandler import MenuHandler
from asn1editor.wxPython.SingleFileDropTarget import SingleFileDropTarget
from asn1editor.wxPython.Styler import Styler
from asn1editor.wxPython.TreeView import TreeView
from asn1editor.wxPython.ViewSelect import ViewType
from asn1editor.wxPython.WxPythonViews import WxPythonView


class MainWindow(wx.Frame, PluginInterface):
    def __init__(self, plugins: typing.Optional[typing.List[Plugin]] = None, title=f'ASN.1 editor {asn1editor.__version__}', enable_load_last=True):
        super(MainWindow, self).__init__(None, title=title)
        self.__title = title

        Environment.load()
        Resources.image_list = ImageList()

        if plugins is not None:
            for plugin in plugins:
                plugin.connect(self)

        self._status_bar = self.CreateStatusBar()

        self._menu_handler = MenuHandler(self, plugins, self.__about_box_content(title, plugins))

        self._menu_handler.build(self.load_spec, self.load_data_from_file, self.save_data_to_file, self._structure_changed)
        self.Bind(wx.EVT_CLOSE, self.close)

        self.SetSize(wx.Size(Environment.settings.get('size', (500, 800))))
        self.Maximize(Environment.settings.get('maximized', True))
        self.SetPosition(wx.Point(Environment.settings.get('position', (0, 0))))
        screen_rect: wx.Rect = self.GetScreenRect()
        center = screen_rect.GetTopLeft() + (screen_rect.GetWidth() // 2, screen_rect.GetHeight() // 2)
        if wx.Display.GetFromPoint(center) == wx.NOT_FOUND:
            self.SetPosition(wx.Point(0, 0))
        self._menu_handler.view_select.selected = ViewType(Environment.settings.get('view', ViewType.TREE.value))
        self._menu_handler.recent = Environment.settings.get('recent', [])
        self._menu_handler.load_last = Environment.settings.get('load_last', True)

        self.__asn1_handler: typing.Optional[ASN1SpecHandler] = None

        self.__model = None
        self.__view = None
        self.__controller = None
        self.__type_name = None
        self.__file_name = None
        self.__tree_view: typing.Optional[TreeView] = None
        self.__content_panel: typing.Optional[wx.ScrolledWindow] = None

        self.__progress_window: typing.Optional[wx.ProgressDialog] = None

        self.SetDropTarget(SingleFileDropTarget(self.__file_dropped))
        self.SetAutoLayout(True)
        self.SetSizer(wx.BoxSizer(wx.HORIZONTAL))

        self.__default_excepthook = sys.excepthook
        # noinspection SpellCheckingInspection
        sys.excepthook = self.__exception_handler

        WxPythonView.structure_changed = None

        if self._menu_handler.load_last and enable_load_last:
            # noinspection PyBroadException
            try:
                self._menu_handler.load_most_recent()
            except Exception:
                pass

    def __file_dropped(self, file_name: str):
        if self.__asn1_handler is not None:
            self.load_data_from_file(file_name)
        else:
            self.load_spec(file_name)

    def load_spec(self, file_name: typing.Optional[str], type_name: typing.Optional[str] = None):
        wx.App.Get().ProcessPendingEvents()
        if file_name is None:
            # Close spec
            self.__asn1_handler = None
            self.__type_name = None

            if self.__view is not None:
                self.__view.realize().destroy()
                self.__content_panel.Destroy()
                self.__view = None
            if self.__tree_view is not None:
                self.__tree_view.destroy()
                self.__tree_view = None
            self.SetTitle(self.__title)
            return

        # Spec file loaded, compile it to show a selection of type names
        if not self.__asn1_handler or file_name not in self.__asn1_handler.get_filename():
            try:
                self.__asn1_handler = ASN1SpecHandler(file_name)
            except FileNotFoundError:
                self.show_message(f'File {file_name} not found', 'Error', PluginInterface.MessageType.ERROR)
                return

        if type_name is None:
            types = self.__asn1_handler.get_types()
            dialog = wx.SingleChoiceDialog(self, 'Select type from ASN.1 file', 'Select type', types)
            try:
                if dialog.ShowModal() == wx.ID_OK:
                    self.__type_name = dialog.GetStringSelection()
            finally:
                dialog.Destroy()
        else:
            self.__type_name = type_name

        if self.__type_name is not None:
            self._menu_handler.add_recent(os.path.abspath(file_name), self.__type_name)

            self._status_bar.SetStatusText(f'Loaded {file_name}')
            self.SetTitle(f'{self.__title} - {file_name}')
            self.__file_name = file_name

            if self.__view is not None:
                self.__view.realize().destroy()
                self.__content_panel.Destroy()
            if self.__tree_view is not None:
                self.__tree_view.destroy()

            styler = Styler(os.path.splitext(file_name)[0] + '.style')

            WxPythonView.structure_changed = lambda x: None

            self.__content_panel = wx.ScrolledWindow(self, style=wx.HSCROLL | wx.VSCROLL)
            self.__content_panel.SetScrollbars(15, 15, 50, 50)
            self.__content_panel.SetAutoLayout(True)
            self.__content_panel.SetSizer(wx.BoxSizer(wx.VERTICAL))

            view_factory = WxPythonViewFactory.WxPythonViewFactory(self.__content_panel, styler)

            self.Freeze()

            self.__view, self.__controller = self.__asn1_handler.create_mvc_for_type(self.__type_name, view_factory)
            self.__tree_view = TreeView(self, self.__content_panel, self.__type_name)

            self.Thaw()

            WxPythonView.structure_changed = self._structure_changed
            self._structure_changed()

            self._menu_handler.enable()

    def _structure_changed(self):
        if self.__type_name is None:
            return

        self.Freeze()

        sizer = self.GetSizer()
        sizer.Clear()
        content_panel_sizer: wx.Sizer = self.__content_panel.GetSizer()
        content_panel_sizer.Clear()

        if self._menu_handler.view_select.selected == ViewType.TREE:

            tree_ctrl = self.__tree_view.get_ctrl(self.__view.realize())

            sizer.Add(tree_ctrl, proportion=1, flag=wx.ALL | wx.EXPAND)
            sizer.Add(self.__content_panel, proportion=2, flag=wx.ALL | wx.EXPAND)
        else:
            self.__tree_view.hide()

            self.__view.realize().set_visible(True, recursive=True)
            content_sizer = self.__view.realize().get_sizer(recursive=True)
            content_panel_sizer.Add(content_sizer, flag=wx.ALL | wx.EXPAND, border=5)

            self.__content_panel.SetSizerAndFit(content_panel_sizer)

            sizer.Add(self.__content_panel, flag=wx.ALL | wx.EXPAND)

        self.SetSizer(sizer)

        self.Refresh()
        self.PostSizeEvent()

        self.Thaw()

    def load_data_from_file(self, file_name: str):
        self.__controller.model_to_view(self.__asn1_handler.load_data_file(file_name))
        self._status_bar.SetStatusText(f'Loaded {file_name} for {self.__type_name}')

    def save_data_to_file(self, file_name: str):
        self.__asn1_handler.save_data_file(file_name, self.__controller.view_to_model())

    def show_data(self, data: bytes, codec: str):
        self.__controller.model_to_view(self.__asn1_handler.get_model_from_data(data, codec))
        self._status_bar.SetStatusText(f'Loaded data for {self.__type_name}')

    def file_picker(self, message: str, wildcard: str, open_: bool) -> typing.Optional[str]:
        def dialog_constructor() -> wx.FileDialog:
            return wx.FileDialog(self, message, wildcard=wildcard,
                                 style=(wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) if open_ else wx.FD_SAVE)

        picker = FilePickerHandler(dialog_constructor, None, not open_)
        picker.on_menu_click(wx.EVT_MENU)
        return picker.filename

    def dir_picker(self, message: str) -> typing.Optional[str]:
        def dialog_constructor() -> wx.DirDialog:
            return wx.DirDialog(self, message,
                                style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)

        picker = FilePickerHandler(dialog_constructor, None)
        picker.on_menu_click(wx.EVT_MENU)
        return picker.filename

    def encode_data(self, codec: str) -> bytes:
        return self.__asn1_handler.get_data_from_model(self.__controller.view_to_model(), codec)

    def get_spec_filename(self) -> str:
        return self.__file_name

    def get_typename(self) -> str:
        return self.__type_name

    def get_spec(self, codec: str):
        return self.__asn1_handler.get_compiled(codec)

    def text_entry(self, message: str, default: typing.Optional[str] = None) -> typing.Optional[str]:
        with wx.TextEntryDialog(self, message) as text_dialog:
            if default is not None:
                text_dialog.SetValue(default)
            if text_dialog.ShowModal() == wx.ID_CANCEL:
                return
            return text_dialog.GetValue()

    def choice_entry(self, message: str, caption: str, choices: typing.List[str], default: typing.Optional[str] = None) -> typing.Optional[str]:
        with wx.SingleChoiceDialog(self, message, caption, choices=choices) as choice_dialog:
            if default is not None:
                try:
                    choice_dialog.SetSelection(choices.index(default))
                except ValueError:
                    pass
            if choice_dialog.ShowModal() == wx.ID_CANCEL:
                return
            return choice_dialog.GetStringSelection()

    def show_status(self, message: str):
        self._status_bar.SetStatusText(message)

    def show_message(self, message: str, caption: str, message_type: PluginInterface.MessageType) -> bool:
        style = wx.CENTER | {PluginInterface.MessageType.WARNING: wx.OK | wx.ICON_WARNING,
                             PluginInterface.MessageType.INFO: wx.OK | wx.ICON_INFORMATION,
                             PluginInterface.MessageType.ERROR: wx.OK | wx.ICON_ERROR,
                             PluginInterface.MessageType.QUESTION: wx.YES_NO | wx.ICON_QUESTION}[message_type]
        ret = wx.MessageBox(message, caption, style=style)
        return True if message_type != PluginInterface.MessageType.QUESTION else wx.YES == ret

    def show_progress(self, message: str, caption: str, max_progress: typing.Optional[int] = None):
        self.__progress_window = wx.ProgressDialog(caption, message, maximum=max_progress if max_progress else 100, style=wx.PD_APP_MODAL | wx.PD_CAN_ABORT)

    def update_progress(self, message: typing.Optional[str] = None, close: bool = False, progress: typing.Optional[int] = None) -> bool:
        running = False
        if self.__progress_window:
            if progress is not None:
                running = self.__progress_window.Update(progress, newmsg=message)[0]
            else:
                running = self.__progress_window.Pulse()[0]
        if close:
            self.__progress_window.Close()
            self.__progress_window = None

        return running

    def get_settings(self) -> dict:
        return Environment.settings.setdefault('Plugin', {})

    def select_view(self, view: ViewType):
        self._menu_handler.view_select.selected = view

    def __exception_handler(self, exc_type, value, trace):
        import traceback
        trace = ''.join(traceback.format_exception(exc_type, value, trace))
        print(trace)

        message = f'{value}\n\n{str(exc_type)}'
        exception_str = f'{message}\n\n{trace}'
        wx.MessageBox(message, 'Error', wx.OK | wx.ICON_ERROR, parent=self)
        self._status_bar.SetStatusText(f'Error: {value}')
        Environment.log_error(exception_str)

    # noinspection PyUnusedLocal
    def close(self, e: wx.Event):
        Environment.settings['size'] = (self.GetSize().Get())
        Environment.settings['maximized'] = self.IsMaximized()
        Environment.settings['position'] = self.GetPosition().Get()
        Environment.settings['view'] = self._menu_handler.view_select.selected.value
        Environment.settings['recent'] = self._menu_handler.recent[:10]
        Environment.settings['load_last'] = self._menu_handler.load_last

        Environment.save()

        sys.excepthook = self.__default_excepthook

        self.Destroy()

    @staticmethod
    def __about_box_content(title: str, plugins: typing.List[Plugin]) -> str:
        import asn1tools

        my_version = f'ASN.1 editor {asn1editor.__version__}'
        if my_version not in title:
            title += f' based on {my_version}'

        plugin_strs = []
        if plugins:
            for plugin in plugins:
                about = plugin.get_about()
                if about:
                    plugin_strs.append(f'{plugin.get_name()}: {about}')

        plugin_str = '\n' + '\n'.join(plugin_strs) + '\n'

        return f'''{title}
{plugin_str}
Published under MIT License

Copyright (c) 2020 Florian Fetz
https://github.com/Futsch1/asn1editor

Based on eerimoq's asn1tools, used in {asn1tools.version.__version__}
https://github.com/eerimoq/asn1tools
'''



from typing import Callable


class OptionalInterface:  # pragma: no cover
    def register_optional_event(self, callback: Callable):
        raise NotImplementedError

    def get_has_value(self) -> bool:
        raise NotImplementedError

    def set_has_value(self, val: bool):
        raise NotImplementedError



__version__ = "0.3.3"
# noinspection SpellCheckingInspection
__author__ = "Florian Fetz"

from . import wxPython
from .Plugin import Plugin
from .PluginInterface import PluginInterface

__all__ = ['wxPython', 'Plugin', 'PluginInterface']



import typing

import wx


class ImageList:
    def __init__(self):
        self.__image_list = wx.ImageList(16, 16, False, 8)
        self.__index = {}

    def add_bitmap(self, bitmap: wx.Bitmap, name: str):
        self.__index[name] = self.__image_list.Add(bitmap)

    def get_bitmap(self, name: str) -> typing.Optional[wx.Bitmap]:
        if name in self.__index:
            return self.__image_list.GetBitmap(self.__index[name])

    def get_index(self, name: str) -> int:
        return self.__index[name]

    def get_image_list(self) -> wx.ImageList:
        return self.__image_list



import enum
import typing

import asn1tools


class PluginInterface:  # pragma: no cover

    class MessageType(enum.Enum):
        INFO = 1
        WARNING = 2
        ERROR = 3
        QUESTION = 4

    def load_spec(self, file_name: str, type_name: typing.Optional[str] = None):
        """
        Loads an ASN.1 specification in the editor.

        :param file_name: File name of the ASN.1 specification
        :param type_name: Optional type name indication. Format: [Module name].[Type name]
        """
        raise NotImplementedError

    def show_data(self, data: bytes, codec: str):
        """
        Decodes and shows data from an ASN.1 encoded byte stream in the editor.

        :param data: Byte stream with ASN.1 encoded data
        :param codec: Codec to use for decoding (reference asn1tools which ones are supported)
        """
        raise NotImplementedError

    def encode_data(self, codec: str) -> bytes:
        """
        Encodes the data currently edited in the editor window with an ASN.1 codec.

        :param codec: Codec to use for encoding (reference asn1tools which ones are supported)

        :return: Encoded data
        """
        raise NotImplementedError

    def get_spec_filename(self) -> typing.Optional[str]:
        """
        :return: The file name of the ASN.1 specification file currently loaded.
        """
        raise NotImplementedError

    def get_spec(self, codec: str) -> asn1tools.compiler.Specification:
        """
        :param codec: Codec to use (reference asn1tools which ones are supported)

        :return: The compiled ASN.1 specification as an object defined by asn1tools.
        """
        raise NotImplementedError

    def get_typename(self) -> typing.Optional[str]:
        """
        :return: The name of the currently used type name in the editor.
        """
        raise NotImplementedError

    def file_picker(self, message: str, wildcard: str, open_: bool) -> typing.Optional[str]:
        """
        Opens a file picker where the user can choose a file name and returns this chosen name.

        :param message: Message to show in the file picker
        :param wildcard: Wildcards to offer the user in order to help him choose the correct file types
        :param open_: Flag if the dialog shall be an open dialog (True) or a save dialog (false)

        :return: Selected file, None if cancelled
        """
        raise NotImplementedError

    def dir_picker(self, message: str) -> typing.Optional[str]:
        """
        Opens a dir picker where the user can choose a directory name and returns this chosen name.

        :param message: Message to show in the dir picker

        :return: Selected directory, None if cancelled
        """
        raise NotImplementedError

    def text_entry(self, message: str, default: typing.Optional[str] = None) -> typing.Optional[str]:
        """
        Queries a text from the user.

        :param message: Message to show to the user
        :param default: Default value for the text entry

        :return: Entered text, None if cancelled
        """
        raise NotImplementedError

    def choice_entry(self, message: str, caption: str, choices: typing.List[str], default: typing.Optional[str] = None) -> typing.Optional[str]:
        """
        Queries the user for a choice of values.

        :param message: Message to show to the user
        :param caption: Message box caption
        :param choices: List of strings for the user to choose from
        :param default: Default selection

        :return: Selected string, None if cancelled
        """
        raise NotImplementedError

    def show_message(self, message: str, caption: str, message_type: MessageType) -> bool:
        """
        Shows a status message.

        :param message: Message to show to the user
        :param caption: Message box caption
        :param message_type: Type of message box

        :return: If type is a question, True means yes and False means No
        """
        raise NotImplementedError

    def show_status(self, message: str):
        """
        Shows a status message.

        :param message: Message to show to the user
        """
        raise NotImplementedError

    def get_settings(self) -> dict:
        """
        Get a settings dictionary for a plugin to store persistent data

        :return: Settings dict
        """
        raise NotImplementedError

    def show_progress(self, message: str, caption: str, max_progress: typing.Optional[int] = None):
        """
        Shows a progress window

        :param message: Message to show to the user
        :param caption: Message box caption
        :param max_progress: Maximum progress value
        """
        raise NotImplementedError

    def update_progress(self, message: typing.Optional[str] = None, close: bool = False, progress: typing.Optional[int] = None) -> bool:
        """
        Updates a shown progress window

        :param message: Updated message to show to the user, optional
        :param close: If set, window is closed, optional
        :param progress: Progress to show, if none, an indeterminate process is shown, optional

        :return: False if progress window has been canceled, otherwise True
        """
        raise NotImplementedError



import typing
from unittest import TestCase

from asn1editor.controller import Controller
from asn1editor.interfaces.BitstringInterface import BitstringInterface
from tests.controller.test_valueBoolControllers import TestOptionalInterface


class TestBitstringInterface(BitstringInterface):

    def __init__(self):
        self.values = []

    def get_values(self) -> typing.List[int]:
        return self.values

    def set_values(self, values: typing.List[int]):
        self.values = values


class TestBitstringController(TestCase):
    def test_add_controller(self):
        with self.assertRaises(Exception):
            root = Controller.RootController('root')
            controller = Controller.BitstringController('test', root, TestBitstringInterface(), None, 5)
            controller.add_controller('test', controller)

    def test_model_to_view(self):
        root = Controller.RootController('root')
        value_interface = TestBitstringInterface()
        optional_interface = TestOptionalInterface()

        controller = Controller.BitstringController('test', root, value_interface, None, 5)
        with self.assertRaises(AssertionError):
            controller.model_to_view({})

        controller = Controller.BitstringController('test', root, value_interface, optional_interface, 6)

        controller.model_to_view({})
        self.assertFalse(optional_interface.val)

        with self.assertRaises(AssertionError):
            controller.model_to_view({'test': (b'\x00', 7)})

        controller.model_to_view({'test': (b'\x33', 6)})
        self.assertListEqual(value_interface.values, [0, 1, 4, 5])

        controller = Controller.BitstringController('test', root, value_interface, optional_interface, 22)

        controller.model_to_view({'test': (b'\x82\x81\x81', 22)})
        self.assertListEqual(value_interface.values, [1, 7, 8, 15, 16])

    def test_view_to_model(self):
        root = Controller.RootController('root')
        value_interface = TestBitstringInterface()
        optional_interface = TestOptionalInterface()

        controller = Controller.BitstringController('test', root, value_interface, optional_interface, 22)

        optional_interface.val = False
        self.assertIsNone(controller.view_to_model())

        optional_interface.val = True
        self.assertEqual(controller.view_to_model(), (b'\x00\x00\x00', 22))

        value_interface.values = [1, 7, 8, 15, 16]
        self.assertEqual(controller.view_to_model(), (b'\x82\x81\x01', 22))



from typing import Tuple, Union, Any

from asn1tools.codecs import constraints_checker
from asn1tools.compiler import oer

from asn1editor.controller.ChoiceInstanceFactory import ChoiceInstanceFactory
from asn1editor.controller.Controller import Controller, RootController
from asn1editor.controller.ControllerFactory import ControllerFactory
from asn1editor.controller.ListInstanceFactory import ListInstanceFactory
from asn1editor.view.AbstractView import AbstractView
from asn1editor.view.AbstractViewFactory import AbstractViewFactory


class ViewControllerFactory(object):
    def __init__(self, view_factory: AbstractViewFactory):
        self._view_factory = view_factory

    def create(self, asn1_type: oer.CompiledType) -> Tuple[AbstractView, Controller]:
        controller = RootController('root')
        view = self.create_view_and_controller(asn1_type.type, asn1_type.constraints_checker.type, controller)

        return view, controller

    def create_view_and_controller(self, type_: oer.Type, checker: constraints_checker.Type, controller: Controller) -> AbstractView:
        if isinstance(type_, oer.Integer):
            return self._number(type_, checker, controller)
        elif isinstance(type_, oer.Boolean):
            return self._bool(type_, controller)
        elif isinstance(type_, oer.Real):
            return self._number(type_, checker, controller)
        elif isinstance(type_, oer.Sequence) or isinstance(type_, oer.Set):
            # noinspection PyTypeChecker
            return self._sequence(type_, checker, controller)
        elif isinstance(type_, oer.SequenceOf) or isinstance(type_, oer.SetOf):
            # noinspection PyTypeChecker
            return self._sequence_of(type_, checker, controller)
        elif isinstance(type_, oer.Choice):
            # noinspection PyTypeChecker
            return self._choice(type_, checker, controller)
        elif isinstance(type_, oer.OctetString):
            # noinspection PyTypeChecker
            return self._hex_string(type_, checker, controller)
        elif type(type_) in [oer.UTF8String, oer.VisibleString, oer.GeneralString, oer.IA5String, oer.ObjectIdentifier]:
            # noinspection PyTypeChecker
            return self._string(type_, checker, controller)
        elif isinstance(type_, oer.Enumerated):
            return self._enumerated(type_, controller)
        elif isinstance(type_, oer.BitString):
            # noinspection PyTypeChecker
            return self._bitstring(type_, controller)
        elif isinstance(type_, oer.Null):
            return self._null(type_, controller)
        elif isinstance(type_, oer.Recursive):
            # noinspection PyUnresolvedReferences
            return self.create_view_and_controller(type_.inner, checker.inner, controller)
        else:
            return self._text(type_, f'ASN.1 type {type_.name} {type_.type_name} not supported')

    def _text(self, type_: oer.Type, text: str) -> AbstractView:
        return self._view_factory.get_text_view(type_.name, text)

    def _null(self, type_: oer.Type, controller: Controller) -> AbstractView:
        ControllerFactory(controller).create_null_controller(type_)
        return self._view_factory.get_text_view(type_.name, "NULL")

    def _number(self, type_: Union[oer.Integer, oer.Real], checker: constraints_checker.Type, controller: Controller) -> AbstractView:
        view, value_interface, optional_interface = self._view_factory.get_number_view(type_.name, type_.optional, self.__get_limit(checker.minimum),
                                                                                       self.__get_limit(checker.maximum), isinstance(type_, oer.Real))

        ControllerFactory(controller).create_value_controller(type_, value_interface, optional_interface, self.__get_limit(checker.minimum))

        return view

    def _sequence(self, type_: oer.Sequence, checker: constraints_checker.Dict, controller: Controller) -> AbstractView:
        view, optional_interface = self._view_factory.get_container_view(type_.name, type_.optional)

        sub_controller = ControllerFactory(controller).create_container_controller(type_, optional_interface)

        for sub_type in type_.root_members + ([] if type_.additions is None else type_.additions):
            view.add_child(self.create_view_and_controller(sub_type, self.__get_member_checker(checker, sub_type.name), sub_controller))

        return view

    def _sequence_of(self, type_: oer.SequenceOf, checker: constraints_checker.List, controller: Controller) -> AbstractView:
        view, value_interface, optional_interface = self._view_factory.get_list_view(type_.name, self.__get_limit(checker.minimum),
                                                                                     self.__get_limit(checker.maximum), type_.optional)

        list_instance_factory = ListInstanceFactory(self._view_factory, view, type_.element_type, checker.element_type)
        ControllerFactory(controller).create_list_controller(type_, value_interface, optional_interface, list_instance_factory,
                                                             self.__get_limit(checker.minimum))

        return view

    def _enumerated(self, type_: oer.Enumerated, controller: Controller) -> AbstractView:
        choices = [str(value) for value in type_.value_to_data.values()]
        view, value_interface, optional_interface = self._view_factory.get_enumerated_view(type_.name, choices, type_.optional)

        ControllerFactory(controller).create_value_controller(type_, value_interface, optional_interface)

        return view

    def _bool(self, type_: oer.Boolean, controller: Controller):
        view, value_interface, optional_interface = self._view_factory.get_boolean_view(type_.name, type_.optional)

        ControllerFactory(controller).create_bool_controller(type_, value_interface, optional_interface)

        return view

    def _string(self, type_: oer.VisibleString, checker: constraints_checker.String, controller: Controller):
        view, value_interface, optional_interface = self._view_factory.get_string_view(type_.name, type_.type_name, type_.optional,
                                                                                       self.__get_limit(checker.minimum), self.__get_limit(checker.maximum))

        ControllerFactory(controller).create_value_controller(type_, value_interface, optional_interface, self.__get_limit(checker.minimum))

        return view

    def _hex_string(self, type_: oer.OctetString, checker: constraints_checker.String, controller: Controller):
        view, value_interface, optional_interface = self._view_factory.get_hex_string_view(type_.name, type_.optional, self.__get_limit(checker.minimum),
                                                                                           self.__get_limit(checker.maximum))

        ControllerFactory(controller).create_value_controller(type_, value_interface, optional_interface, self.__get_limit(checker.minimum))

        return view

    def _bitstring(self, type_: oer.BitString, controller: Controller):
        if type_.number_of_bits is None:
            view, value_interface, optional_interface = self._view_factory.get_hex_string_view(type_.name, type_.optional, None, None)

            ControllerFactory(controller).create_value_controller(type_, value_interface, optional_interface, None)
        else:
            view, value_interface, optional_interface = self._view_factory.get_bitstring_view(type_.name, type_.number_of_bits, type_.named_bits,
                                                                                              type_.optional)

            ControllerFactory(controller).create_bitstring_controller(type_, value_interface, optional_interface)

        return view

    def _choice(self, type_: oer.Choice, checker: constraints_checker.Choice, controller: Controller):
        choices = [member.name for member in type_.members]
        view, value_interface, optional_interface = self._view_factory.get_choice_view(type_.name, choices, type_.optional)

        members = {member.name: member for member in type_.members}
        checkers = {member.name: member for member in checker.members}

        choice_instance_factory = ChoiceInstanceFactory(self._view_factory, view, members, checkers)
        ControllerFactory(controller).create_choice_controller(type_, value_interface, optional_interface, choice_instance_factory)

        return view

    @staticmethod
    def __get_member_checker(checker: constraints_checker.Dict, name: str) -> constraints_checker.Type:
        for member in checker.members:
            if member.name == name:
                return member

    @staticmethod
    def __get_limit(limit: Any):
        return None if limit in ['MIN', 'MAX'] else limit



from typing import Dict, Optional

from asn1tools.codecs import oer, constraints_checker

from asn1editor.controller.Controller import Controller
from asn1editor.view.AbstractView import AbstractView, ChoiceView
from asn1editor.view.AbstractViewFactory import AbstractViewFactory


class ChoiceInstanceFactory:
    def __init__(self, view_factory: AbstractViewFactory, choice_view: ChoiceView, members: Dict[str, oer.Type], checkers: Dict[str, constraints_checker.Type]):
        self._view_factory = view_factory
        self._choice_view = choice_view
        self._members = members
        self._checkers = checkers

        self._content_view: Optional[AbstractView] = None
        self._member = None

    def create(self, member: str, parent: Controller):
        from asn1editor.ViewControllerFactory import ViewControllerFactory

        if member != self._member:
            view_factory = ViewControllerFactory(self._view_factory)

            self._content_view = view_factory.create_view_and_controller(self._members[member], self._checkers[member], parent)
            self._member = member
            self._choice_view.set_view(self._content_view)



from typing import Dict

from asn1tools.codecs import oer, constraints_checker

from asn1editor.controller.Controller import Controller
from asn1editor.view.AbstractView import ListView, AbstractView
from asn1editor.view.AbstractViewFactory import AbstractViewFactory


class ListInstanceFactory:
    def __init__(self, view_factory: AbstractViewFactory, list_view: ListView, _type: oer.Type, checker: constraints_checker.Type):
        self._view_factory = view_factory
        self._list_view = list_view
        self._type = _type
        self._checker = checker

        self.content_views: Dict[int, AbstractView] = {}

    def create(self, instance: int, parent: Controller):
        from asn1editor.ViewControllerFactory import ViewControllerFactory

        view_factory = ViewControllerFactory(self._view_factory)
        self._type.name = f'Element {instance}'
        self.content_views[instance] = view_factory.create_view_and_controller(self._type, self._checker, parent)

        self._list_view.add(self.content_views[instance])

    def destroy(self, instance: int):
        self._list_view.remove(self.content_views[instance])

        del self.content_views[instance]
